import logging
import threading
import time


class Daemon:
    """Class to run a periodical task in another thread

    Usage
    -----
    - Subclass Daemon
    - call super().__init__() in yours
    - Overwrite on_process method with yours
    - Add logic you want to run inside on_process
    - Be sure your methods are safe-thread to avoid race condition
    - self.lock is available for lock.acquire / your_logic / lock.release
    - call start() method to keep running on_process in a new thread
    - call stop() to finish the thread

    - check minibone.sample_clock.py if you want to learn how to use it

    Usage callback mode
    -------------------
    - Instance Daemon by passing a callable
    - Add logic to your callable method
    - Be sure your callable is safe-thread to avoid race condition
    - call start() method to keep running callable in a new thread
    - call stop() to finish the thread

    - check minibone.sample_clock_callback.py if you want to learn how to use it

    Notes:
    ------
    start() must be called only once
    """

    def __init__(
        self,
        name: str = None,
        interval: int = 60,
        sleep: float = 0.5,
        callback=None,
        iter: int = -1,
        daemon: bool = True,
        **kwargs,
    ):
        """
        Arguments
        ---------
        name        str         name for this thread

        interval    int         Number of interval seconds to run on_process.
                                Must be >= 0

        sleep       int         Number of seconds to sleep on each interation when iddle.
                                Must be >= 0 and <= 1. Set to Zero to do not sleep
                                Sleep happends after calling on_process/callback

        callback    callable    [Optional] A callable object to be called instead of on_process
                                Default None.

        iter        int         How many times to run this task. iter must be >= 1 or -1
                                -1 runs forever until stopped

        daemon      bool        True to set the Thread as a daemon, False otherwise

        kwargs                  Additional params you need to pass

        Notes
        -----
        sleep will block the thread, so if stop is called it will wait until sleep is done.
        sleep was implemented as a convenient way to avoid? it does not get resources hungry
        when in iddle state inside the bucle iterating/waiting for something to do

        Thumb of usage for sleep:
        Set to 0.01 if on_process is high priority
        """
        assert not name or isinstance(name, str)
        assert isinstance(interval, (float, int)) and interval >= 0
        assert isinstance(sleep, (float, int)) and sleep >= 0 and sleep <= 1
        assert not callback or callable(callback)
        assert isinstance(iter, int) and (iter == -1 or iter >= 1)
        assert isinstance(daemon, bool)
        self._logger = logging.getLogger(__class__.__name__)

        self.lock = threading.Lock()
        self._stopping = False

        self._name = name
        self._interval = interval
        self._sleep = sleep
        self._check = 0
        self._iter = iter
        self._count = 0

        self._callback = callback

        self._process = threading.Thread(
            name=self._name, target=self._do_process, kwargs=kwargs, daemon=True if daemon else None
        )

    def on_process(self):
        """Method to be called on each interation.
        Overwrite it with your logic if a callback is not set

        Do not forget to make your code safe-thread using lock.acquire and lock.release
        """
        pass

    def _do_process(self, **kwargs):
        while True:
            if self._stopping:
                return

            epoch = time.time()
            if epoch > self._check:
                self._check = epoch + self._interval

                if self._callback:
                    self._callback(**kwargs)
                else:
                    self.on_process(**kwargs)

                if self._iter > 0:
                    self._count += 1
                    if self._count >= self._iter:
                        return

            if self._sleep > 0:
                time.sleep(self._sleep)

    def start(self):
        """Start running on_process/callback periodically"""

        self._process.start()

        self._logger.debug(
            "started %s task at interval: %.2f sleep: %.2f iterate: %d",
            self._name,
            self._interval,
            self._sleep,
            self._iter,
        )

    def stop(self):
        """Stop interating on on_process/callback and exit this thread"""
        self.lock.acquire()
        self._stopping = True
        self.lock.release()
        self._process.join()

        self._logger.debug(
            "stopping %s task at interval: %.2f sleep: %.2f iterate: %d",
            self._name,
            self._interval,
            self._sleep,
            self._iter,
        )



"""This is a simple clock sample.

It prints current time each second using a SubClass of Daemon
"""

import time
from datetime import datetime

from daemon import Daemon


class Clock(Daemon):
    def __init__(self):
        super().__init__(name="Clock", interval=1, sleep=0.01)

    def on_process(self):
        print(datetime.now().strftime("%Y-%m-%d:%H:%M:%S"))


if __name__ == "__main__":
    try:
        print("Sample clock subclassing Daemon")
        print("Press ctrl+c to exit")
        clock = Clock()
        clock.start()

        while True:
            print("I am going to sleep. Do not disturb please")
            sleeping = 15
            time.sleep(sleeping)
            # clock will be running in the background in another thread
            print("I awoke after a long sleep of {%d} seconds in the main thread" % sleeping)

    except KeyboardInterrupt:
        print("Caught keyboard interrupt, exiting")

    except Exception as e:
        print("Error %s" % e)

    finally:
        clock.stop()
        print("Doing Shutdown")



import asyncio
import unittest
from pathlib import Path

from minibone.html_base import HTMLBase


class TestHTMLBase(unittest.TestCase):
    def test_htmlbase(self):
        pathsnippets = Path("./snippets")
        pathsnippets.mkdir(parents=True, exist_ok=True)

        mapping = {"user": "Max"}
        snippet = "<div>Hello ${user}</div>"
        html = "<!DOCTYPE html><html lang='es-ES'><head><title>${title}</title></head><body>${account}</body>"
        toml = """
        [page]
        html_file = 'index.html'
        title = 'HTMLBase'

        [account]
        user = 'John'        
        """

        with open("./snippets/account.html", "tw") as f:
            f.write(snippet)

        with open("./snippets/account.txt", "tw") as f:
            f.write(snippet)

        with open("./index.html", "tw") as f:
            f.write(html)

        with open("./index.toml", "tw") as f:
            f.write(toml)

        htmlbase = HTMLBase(snippets_path="./snippets")
        self.assertEqual(htmlbase.render(snippet, mapping), "<div>Hello Max</div>")
        self.assertEqual(asyncio.run(htmlbase.aio_file("./snippets/account.html")), "<div>Hello ${user}</div>")
        self.assertEqual(
            asyncio.run(htmlbase.aiofrom_toml("index.toml")),
            "<!DOCTYPE html><html lang='es-ES'><head><title>HTMLBase</title></head><body><div>Hello John</div></body>",
        )

        htmlbase = HTMLBase(snippets_path="./snippets", ext="txt")
        self.assertEqual(
            asyncio.run(htmlbase.aiofrom_toml("index.toml")),
            "<!DOCTYPE html><html lang='es-ES'><head><title>HTMLBase</title></head><body><div>Hello John</div></body>",
        )

        files = ["./index.toml", "./index.html", "./snippets/account.html", "./snippets/account.txt"]
        for file in files:
            p = Path(file)
            p.unlink(missing_ok=True)
        pathsnippets.rmdir()


if __name__ == "__main__":
    unittest.main()



import asyncio
import logging
import random
import threading
import time
from collections import deque
from enum import Enum

from minibone.daemon import Daemon


class TypeQueue(Enum):
    """Emum to define Type of queues

    FIFO = First in, first out
    LIFO = Last in, first out
    """

    FIFO = "FIFO"
    LIFO = "LIFO"


class PARThreads(Daemon):
    """Class to run parallel tasks in threads.

     This is suitable for I/O bounded tasks such as working with files or fetching web pages.
     For compute or CPU bounded tasks use PARProcesses instead.

    Usage
     -----
     - Create a PARThreads instance
     - call start()
     - call queue to add all your concurrent tasks and store the UID
     - call get(UID) to retrieve the callback's result
     - continue your queue / get loop
     - call stop at the end

     Example
     -------
     from minibone.parallel_threads import PARThreads
     import time

     def reporter(message: str):
         # to emulate a busy action
         time.sleep(1)
         return message

     t = PARThreads()
     t.start()

     # tasks will be run on its own CPU process
     uid1 = t.queue(reporter, message="Say")
     uid2 = t.queue(reporter, message="something")

     res1 = t.get(uid1)
     res2 = t.get(uid2)

     t.stop()

     print(res1, res2)
    """

    def __init__(
        self,
        type_queue: TypeQueue = TypeQueue.FIFO,
        max_threads=10,
        interval: int = 0,
        sleep: int = 0.005,
        daemon: bool = True,
    ):
        """
        Arguments
        ---------
        type_queue:     TypeQueue       Type of queue desired (FIFO / LIFO)
        max_threads:    int             Maximum number of threads to run in parallel
        daemon:         bool            True to run as a daemon, False otherwise
        """
        assert isinstance(type_queue, TypeQueue)
        assert isinstance(max_threads, int) and max_threads > 0
        assert isinstance(daemon, bool)
        super().__init__(name="PoolThreads", interval=interval, sleep=sleep, daemon=daemon)

        self._logger = logging.getLogger(__class__.__name__)

        self.type_queue = type_queue
        self._max_threads = max_threads
        self._current_task = 0

        self._queue = deque()
        self._results = {}

        random.seed(time.time())
        self._letters = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"

    def _uid(self, cmd: str) -> str:
        """Return an unique UID

        Arguments
        ---------
        cmd:    str     String to prefix UID
        """
        assert isinstance(cmd, str)
        return "{}{}_{}".format(cmd, self._letters[random.randint(0, len(self._letters) - 1)], time.time())

    def _lookfor_result(self, uid: str) -> object:
        """Search for a result"""
        assert isinstance(uid, str)
        self.lock.acquire()
        resp = self._results.pop(uid, "_no_in_dict_")
        self.lock.release()
        return resp

    def get(self, uid: str, timeout=-1) -> object | None:
        """Return the response mapped to UID

        Arguments
        ---------
        uid:        str     Unique identifier of the expected result
        timeout:    int     Set the number of seconds to wait for a result.
                            Set to zero or negative to wait forever
        """
        assert isinstance(uid, str)
        assert isinstance(timeout, (int, float))

        notimeout = True if timeout <= 0 else False
        epoch = time.time()
        future_epoch = time.time() + timeout
        while notimeout or epoch < future_epoch:
            resp = self._lookfor_result(uid)
            if resp != "_no_in_dict_":
                return resp

            time.sleep(0.01)

        return None

    async def aioget(self, uid: str, timeout=-1) -> object | None:
        """Return the response mapped to UID in async mode

        Arguments
        ---------
        uid:        str     Unique identifier of the expected result
        timeout:    int     Set the number of seconds to wait for a result.
                            Set to zero or negative to wait forever
        """
        assert isinstance(uid, str)
        assert isinstance(timeout, (int, float))

        notimeout = True if timeout <= 0 else False
        epoch = time.time()
        future_epoch = time.time() + timeout
        while notimeout or epoch < future_epoch:
            resp = self._lookfor_result(uid)
            if resp != "_no_in_dict_":
                return resp

            await asyncio.sleep(0.01)

        return None

    def queue(self, callback, cmd: str = "UID", keep: bool = True, **kwargs) -> str:
        """Queue a task and return an UID to get the result later

        Arguments
        ---------
        callback:   object  A callable object
        cmd:        str     Any command to prefix the UID to return
        keep:       bool    Set to True to keep the callback's result in memory
                            Set to False to forget it. Usefull for callbacks that returns nothing
        kwargs:     dict    a dict with key/value parameters

        Notes
        -----
        If keep is set, do not forget to call get(uid) to free memory
        """
        assert callable(callable)
        assert isinstance(keep, bool)
        assert not kwargs or isinstance(kwargs, dict)

        if not kwargs:
            kwargs = dict()

        uid = self._uid(cmd)
        item = {"uid": uid, "callback": callback, "keep": keep, "kwargs": kwargs}

        self.lock.acquire()
        if self.type_queue == TypeQueue.LIFO:
            self._queue.appendleft(item)
        else:
            self._queue.append(item)
        self.lock.release()

        self._logger.debug("Task %s queued", uid)
        return uid

    def _process_task(self, uid, callback, keep, kwargs):
        """Create a thread to run a task"""
        data = callback(**kwargs)

        self.lock.acquire()
        self._current_task -= 1
        if keep:
            self._results[uid] = data
        self.lock.release()
        self._logger.debug("Task %s done. Parallel tasks: %d", uid, self._current_task)

    def on_process(self):
        while len(self._queue) > 0 and self._current_task < self._max_threads:
            self.lock.acquire()
            item = self._queue.popleft()
            self._current_task += 1
            self.lock.release()
            task = threading.Thread(target=self._process_task, kwargs=item)
            task.start()
            self._logger.debug("Task %s started", item["uid"])



import logging
from collections import deque

from minibone.config import FORMAT, Config
from minibone.daemon import Daemon


class Storing(Daemon):
    """Periodical background task to save objects to file

    This is a queue and forget task to save files

    Usage
    -----
    - create an instance of this class
    - call start()
    - add objects to queue
    - call stop() to stop this thread

    Example
    --------
    from minibone.storing import Storing
    task = Storing()
    task.start()
    data = {"key1": "val1", "key2": "val2"}
    task.to_json(data)
    # data will be saved to json format in another thread
    # another code
    # call stop at the end of your program
    task.stop()
    """

    @classmethod
    def json_from_file(cls, pathfile: str) -> dict:
        """This is an alias to Storing.from_json"""
        _logger = logging.getLogger(__class__.__name__)
        _logger.warning("Use from_json instead.  This method will be deprecated at future")
        return cls.from_json(pathfile)

    @classmethod
    def from_json(cls, filepath: str) -> dict:
        """Return a JSON object from a file"""
        return Config.from_file(format=FORMAT.JSON, filepath=filepath)

    def __init__(self, chunks: int = 5, interval: int = 30):
        """
        Arguments
        ---------
        chunks:     int     Max number of objects to process per iteration
        """
        assert isinstance(chunks, int)
        assert isinstance(interval, int)
        super().__init__(name="storing", interval=interval)
        self._logger = logging.getLogger(__class__.__name__)

        # maximum number of items to process from the queue
        self._chunks = chunks
        self._queue = deque()

    def json_to_file(self, path: str, filename: str, data: dict | list):
        """This is an alias to Storing.to_json"""
        self._logger.warning("Use to_json instead.  This method will be deprecated at future")
        self.to_json(path, filename, data)

    def to_json(self, path: str, filename: str, data: dict | list):
        """Queue a JSON object to be stored in path

        Arguments
        ---------
        path        str         The directory to store json into. Do not add trailing slash /
        filename:   str         The filename to use for saving data
        data:       dict|list   A JSON dict or list to be stored
        """
        assert isinstance(path, str)
        assert isinstance(filename, str)
        assert isinstance(data, (dict, list))

        item = {"format": FORMAT.JSON, "path": path, "file": filename, "data": data}
        self._queue.append(item)
        self._logger.info("{}/{} aded to queue".format(path, filename))

    def on_process(self):
        if len(self._queue) == 0:
            return

        for i in range(min(len(self._queue), self._chunks)):
            d = self._queue.popleft()
            filepath = "{}/{}".format(d["path"], d["file"])
            Config.to_file(format=FORMAT.JSON, filepath=filepath, data=d["data"])



import asyncio
import unittest
from pathlib import Path

from minibone.config import Config


class TestConfig(unittest.TestCase):
    def test_settings(self):
        settings = {"setting1": "value1", "setting2": 2, "setting3": True}
        cfg = Config(settings=settings, filepath=None)

        self.assertEqual(cfg.get("setting1", None), "value1")
        self.assertEqual(cfg.get("setting10", None), None)
        self.assertEqual(cfg.get("setting2", None), 2)
        self.assertEqual(cfg.get("setting3", None), True)

        cfg.remove("setting1")
        cfg.add("setting3", False)

        self.assertEqual(cfg.get("setting1", None), None)
        self.assertEqual(cfg.get("setting3", None), False)

        self.assertEqual(cfg.merge({}, {}), {})
        self.assertEqual(cfg.merge(defaults={"x": 1}), {"x": 1})
        self.assertEqual(cfg.merge(settings={"x": 1}), {"x": 1})
        self.assertEqual(cfg.merge(defaults={"x": 1}, settings={"y": 2}), {"x": 1, "y": 2})
        self.assertEqual(cfg.merge(defaults={"z": 1}, settings={"z": 4}), {"z": 4})

        cfgs = []
        files = ["config.toml", "config.yaml", "config.json", "aconfig.toml", "aconfig.yaml", "aconfig.json"]
        for file in files:
            cfgs.append(Config(settings=cfg, filepath=file))

        cfgs[0].to_toml()
        cfgs[1].to_yaml()
        cfgs[2].to_json()

        self.assertEqual(cfgs[0].from_toml(cfgs[0].filepath), cfg)
        self.assertEqual(cfgs[1].from_yaml(cfgs[1].filepath), cfg)
        self.assertEqual(cfgs[2].from_json(cfgs[2].filepath), cfg)

        asyncio.run(cfgs[3].aioto_toml())
        asyncio.run(cfgs[4].aioto_yaml())
        asyncio.run(cfgs[5].aioto_json())

        self.assertEqual(asyncio.run(cfgs[3].aiofrom_toml(cfgs[3].filepath)), cfg)
        self.assertEqual(asyncio.run(cfgs[4].aiofrom_yaml(cfgs[4].filepath)), cfg)
        self.assertEqual(asyncio.run(cfgs[5].aiofrom_json(cfgs[5].filepath)), cfg)

        for file in files:
            p = Path(file)
            p.unlink(missing_ok=True)


if __name__ == "__main__":
    unittest.main()



import logging
from enum import Enum

import requests

from minibone.parallel_threads import PARThreads


class Verbs(Enum):
    POST = "POST"
    GET = "GET"


class HTTPt:
    """
    Fetches HTTP requests in multi-threading mode
    Each request is queued then procesed in FIFO (first in first out) order using a thread for each one

    Usage
    -----

    from minibone.parallel_threads import PARThreads

    worker = PARThreads()
    client = HTTPt(worker)
    worker.start()

    # this could be in one thread
    uid1 = client.queue_get(url="https://httpbin.org/ip", cmd="test")

    # and this another could be in onether thread
    uid2 = client.queue_get(url="https://httpbin.org/headers", cmd="test")


    res1 = client.read_resp(uid1)
    res2 = client.read_resp(uid2)

    print(res1)
    print(res2)

    worker.stop()

    NOTES
    -----
    Each response may take different time to fetch, so latest queued requests could be gotten first
    This class is great for multi-threaded setups
    """

    def __init__(self, worker: PARThreads, timeout: int = 5):
        """
        Arguments
        ---------
        worker:     object      A PARThreads object
        timeout:    int         Time to wait for an response before giving up

        Notes:
        ------
        Pay attention this class do not call worker.start()
        """
        assert isinstance(worker, PARThreads)
        assert isinstance(timeout, int) and timeout > 0
        self._logger = logging.getLogger(__class__.__name__)

        self._timeout = timeout
        self._worker = worker

        requests.packages.urllib3.disable_warnings()
        # not shortcuts, read this
        # https://shuaib.org/technical-guide/resolving-python-requests-tls-ssl-certificate-verification-errors/

        # TODO take a look to https://netnut.io/httpx-vs-requests
        self.fetcher = requests.Session()
        self.fetcher.headers.update({
            "User-Agent": "AppleWebKit/537.36 (KHTML, like Gecko) Chrome/121.0.0.0 Safari/537.36",
            "X-Forwarded-For": "'\"\\--",
        })

    def _queue_request(self, verb: Verbs, url: str, cmd: str = "UID", **kwargs) -> str:
        """Add a request to the queue and return an UID to retrieve the response whit read_resp

        Arguments
        ---------
        verb:       Verbs   A valid httpt.Verbs verb value
        url:        str     The url to request. Include schema http/https
        cmd         str     Arbitrary string to set as command for this request
        payload     dict    Can be None. A dict or list with parameters for the payload

        Notes:
        ------
        The UID to return is prefixed with the value of the cmd argument if provided
        """
        assert isinstance(verb, Verbs)
        assert isinstance(url, str)

        if not kwargs:
            kwargs = dict()
        kwargs["url"] = url

        if verb == Verbs.GET:
            uid = self._worker.queue(self._get, cmd=cmd, **kwargs)
        else:
            uid = self._worker.queue(self._post, cmd=cmd, **kwargs)

        self._logger.debug("queue_request %s %s", verb, url)

        return uid

    def queue_get(self, url: str, cmd: str = "get", params: dict = None) -> str:
        """Add a get request to the queue and return an UID to retrieve the response whit read_resp

        Arguments
        ---------
        url:        str     The url to get. Include schema http/https
        cmd         str     Arbitrary prefixt string to set as command for this request
                            to avoid returning a duplicated UID (returned UID is prefix_timestamp_here)

        Notes:
        ------
        The UID to return is prefixed with the value of the cmd argument if provided
        """
        return self._queue_request(verb=Verbs.GET, url=url, cmd=cmd, params=params)

    def queue_post(self, url: str, cmd: str = "post", payload: dict = None, is_json: bool = True) -> str:
        """Add a post request to the queue and return an UID to retrieve the response whit read_resp

        Arguments
        ---------
        url:        str     The url to post. Include schema http/https
        cmd         str     Arbitrary string to set as command for this request
        payload     dict    Can be None. A dict or list with parameters for the payload
        is_json     bool    Set to True to parse result as JSON, otherwise to parse as text

        Notes:
        ------
        The UID to return is prefixed with the value of the cmd argument if provided
        """
        return self._queue_request(verb=Verbs.POST, url=url, cmd=cmd, payload=payload, is_json=is_json)

    def read_resp(self, uid: str) -> object | None:
        """Return the response for the UID (json|text) or None if not found or it got a timeout

        Argument
        --------
        uid:    str     Unique identier returned by queue_post or queue_get
        """
        assert isinstance(uid, str)
        return self._worker.get(uid=uid, timeout=self._timeout)

    async def aioread_resp(self, uid: str) -> object | None:
        """Return the response for the UID (json|text) or None if not found or it got a timeout

        Argument
        --------
        uid:    str     Unique identier returned by queue_post or queue_get
        """
        assert isinstance(uid, str)
        return await self._worker.aioget(uid=uid, timeout=self._timeout)

    def _get(self, url: str, params: dict = None) -> str:
        assert isinstance(url, str)
        assert not params or isinstance(params, dict)
        self._logger.debug("_get %s", url)

        if not params:
            params= dict()

        resp = None
        try:
            r = self.fetcher.get(url, timeout=self._timeout, verify=False, **params)
            if r.status_code == 200:
                resp = r.text
            else:
                self._logger.warning("Got %s for %s", r.status_code, url)
            r.close()

        except requests.exceptions.ProxyError as e:
            self._logger.error("Proxy error: %s", e)

        except requests.exceptions.RequestException as e:
            self._logger.error("Request error: %s", e)

        except Exception as e:
            self._logger.error("_get error: Url %s %s", url, e)

        return resp

    def _post(self, url: str, payload: dict = None, is_json: bool = True) -> dict | str:
        assert isinstance(url, str)
        assert not payload or isinstance(payload, dict)
        assert isinstance(is_json, bool)
        self._logger.debug("_post %s %s is_json %s", url, payload, is_json)

        if not payload:
            payload = dict()

        resp = None
        try:
            r = self.fetcher.post(url, data=payload, timeout=self._timeout, verify=False)
            if r.status_code == 200 and not r.is_permanent_redirect and not r.is_redirect:
                if is_json:
                    resp = r.json()
                else:
                    resp = r.text
            else:
                self._logger.warning("Got %s or redirect for %s", r.status_code, url)

            r.close()

        except requests.exceptions.JSONDecodeError as e:
            self._logger.error("Expecting a JSON, but other kind of content gotten %s", e)

        except requests.exceptions.ProxyError as e:
            self._logger.error("Proxy error: %s", e)

        except requests.exceptions.RequestException as e:
            self._logger.error("Request error: %s", e)

        except Exception as e:
            self._logger.error("_post error: Url %s Params %s %s", url, payload, e)

        return resp



"""This is a simple clock sample.

It prints current time each second using a Daemon callback
"""

import time
from datetime import datetime

from daemon import Daemon


def callback():
    print(datetime.now().strftime("%Y-%m-%d:%H:%M:%S"))


if __name__ == "__main__":
    try:
        print("Sample clock using Daemon callback for the task")
        print("Press ctrl+c to exit")
        clock = Daemon(name="Clock", interval=1, sleep=0.01, callback=callback)
        clock.start()

        while True:
            print("I am going to sleep. Do not disturb please")
            sleeping = 15
            time.sleep(sleeping)
            # clock will be running in the background in another thread
            print("I awoke after a long sleep of {%d} seconds in the main thread" % sleeping)

    except KeyboardInterrupt:
        print("Caught keyboard interrupt, exiting")

    except Exception as e:
        print("Error %s" % e)

    finally:
        clock.stop()
        print("Doing Shutdown")



import json
import unittest

from minibone.httpt import HTTPt
from minibone.parallel_threads import PARThreads


class TestHTTPt(unittest.TestCase):
    def test_httpt(self):
        worker = PARThreads()
        worker.start()

        client = HTTPt(worker=worker)

        params = {"params": {"foo": "bar"}}

        # this could be in one thread
        uid1 = client.queue_get(url="https://httpbin.org/anything", params=params)

        # and this another could be in onether thread
        uid2 = client.queue_post(url="https://httpbin.org/post")

        resp1 = json.loads(client.read_resp(uid1))
        resp2 = client.read_resp(uid2)

        worker.stop()

        self.assertEqual(resp1["args"]["foo"], "bar")
        self.assertEqual(resp1["url"], "https://httpbin.org/anything?foo=bar")
        self.assertEqual(resp2["url"], "https://httpbin.org/post")


if __name__ == "__main__":
    unittest.main()



import email.utils as utils
import logging
import smtplib
from email.message import EmailMessage

from minibone.daemon import Daemon


class Emailer(Daemon):
    """Class to send emails using a SMTP server

    This class allows to send emails using a queue in a thread as underground task.
    The queue will be processed as FIFO (first in first out)
    Currently only text and html content are supported (not attachments)

    Usage:
    ------
    intance Emailer, call start, call queue to send emails, call stop at the end of your program
    note stop finishes until all queued emails were sent

    Example
    -------
    import Emailer
    import timer

    # use your own server configuration
    emailer = Emailer(host, port, ssl=True, username="user", password="1234")
    emailer.start()
    emailer.queue(
        from_address="me@domain.com",
        to="you@domain.com",
        subject="Notification",
        content_txt="This is a text notification"
        content_html="This is a <b>html</b> notification"
    )

    # your logic

    # sleep to simulate more logic, emails were send in the background in the meantime
    time.sleep(20)

    emailer.stop()
    """

    def __init__(self, host: str, port: int, ssl: bool, username: str = None, password: str = None):
        """
        Arguments
        ---------
        host:       str     host to conenct to
        port:       int     the port number to connect to
        ssl:        bool    True to use SSL encryption (must be supported by the server)
                            Set to False for plain connection (And may God have mercy on your poor soul!)
        username:   str     The username to login to the server
        password:   str     The password to login to the server
        """
        assert isinstance(host, str)
        assert isinstance(port, int)
        assert isinstance(ssl, bool)
        assert not username or isinstance(username, str)
        assert not password or isinstance(password, str)

        super().__init__(name="Emailer", interval=1)

        self._logger = logging.getLogger(__class__.__name__)

        self._host = host
        self._port = port
        self._ssl = ssl
        self._username = username
        self._password = password

        self._queue = []
        self._timeout = 10

    @property
    def host(self) -> str:
        """The host to connect to"""
        return self._host

    @property
    def port(self) -> int:
        """The port to connect to"""
        return self._port

    @property
    def ssl(self) -> bool:
        """The ssl flag using to connect to (True/False)"""
        return self._ssl

    @property
    def queued(self) -> int:
        """Number of emails pending to be delivered"""
        return len(self._queue)

    def queue(
        self,
        from_address: str,
        to: str | list,
        subject: str,
        content_txt: str = None,
        content_html: str = None,
        cc: str = None,
        bcc: str = None,
        replyto: str = None,
    ):
        """Add a new email to the queue to be delivered

        Arguments
        ---------
        from_address:   str     Address to use as email's from
        to:             str     Address to send email to (can be a list of str addresses)
        subject:        str     Email's subject
        content_txt:    str     Emails's content in text format
        content_html:   str     Emails's content in html format
        cc:             str     Address to cc this email to (can be a list of str addresses)
        bcc:            str     Address to bcc this email to (can be a list of str addresses)
        replyto:        str     Addres to get reply if diferrent from from_address
        """
        assert isinstance(from_address, str)
        assert isinstance(to, (str, list))
        assert isinstance(subject, str)
        assert not content_txt or isinstance(content_txt, str)
        assert not content_html or isinstance(content_html, str)
        assert not cc or isinstance(cc, (str, list))
        assert not bcc or isinstance(bcc, (str, list))
        assert not replyto or isinstance(replyto, str)

        if isinstance(to, str):
            to = [to]

        if isinstance(cc, str):
            cc = [cc]

        if isinstance(bcc, str):
            bcc = [bcc]

        # RFC 5322                Internet Message Format             October 2008
        # https://datatracker.ietf.org/doc/html/rfc5322.html
        #
        # fields =   *(trace
        #     *optional-field /
        #     *(resent-date /
        #     resent-from /
        #     resent-sender /
        #     resent-to /
        #     resent-cc /
        #     resent-bcc /
        #     resent-msg-id))
        # *(orig-date /
        # from /
        # sender /
        # reply-to /
        # to /
        # cc /
        # bcc /
        # message-id /
        # in-reply-to /
        # references /
        # subject /
        # comments /
        # keywords /
        # optional-field)

        self.lock.acquire()
        self._queue.append({
            "from": from_address,
            "to": to,
            "subject": subject,
            "text": content_txt,
            "html": content_html,
            "cc": cc,
            "bcc": bcc,
            "replyto": replyto,
        })
        self.lock.release()

    def on_process(self):
        self.lock.acquire()

        if len(self._queue) > 0:
            item = self._queue[0]
            try:
                msg = EmailMessage()

                msg["Date"] = utils.formatdate()
                msg["From"] = item["from"]
                msg["To"] = item["to"]
                msg["Subject"] = item["subject"]

                if item["text"]:
                    msg.set_content(item["text"])
                if item["html"]:
                    msg.add_alternative(item["html"], subtype="html")

                if item["cc"]:
                    msg["Cc"] = item["cc"]

                if item["bcc"]:
                    msg["Bcc"] = item["bcc"]

                if item["replyto"]:
                    msg["Reply-To"] = item["replyto"]

                if self.ssl:
                    s = smtplib.SMTP_SSL(host=self.host, port=self.port, timeout=self._timeout)
                else:
                    s = smtplib.SMTP(host=self.host, port=self.port, timeout=self._timeout)

                if self._username or self._password:
                    s.login(user=self._username, password=self._password)

                s.send_message(msg)
                s.quit()

                item = self._queue.pop(0)

                self._logger.info("Dispatched [{}] to {}".format(item["subject"], item["to"]))

            except Exception as e:
                self._logger.error("Error sending email to %s  %s", item["to"], e)

        self.lock.release()

    def stop(self):
        super().stop()

        if len(self._queue) > 0:
            self._logger.info("Dispaching %d pending emails before stop", len(self._queue))
            while len(self._queue) > 0:
                self._do_process()
                self._logger.info("Dispaching %d pending emails before stop", len(self._queue))



import logging
import logging.handlers
import time


def setup_log(file: str = None, level: str | int = logging.INFO):
    """Setup log using handlers to support file rotation

    Arguments
    ---------
    file:   str     Full path and name of the file.  Set to None to log to stderr instead
    level:  str     The log level. [DEBUG, INFO, WARNING, ERROR, CRITICAL]
    """
    assert not file or isinstance(file, str)
    assert isinstance(level, (int, str))

    format = "%(asctime)s UTC [%(levelname)s] %(name)s: %(message)s"
    datefmt = "%Y-%m-%d %H:%M:%S"

    if file:
        formatter = logging.Formatter(fmt=format, datefmt=datefmt)
        # date time in GMT/UTC
        formatter.converter = time.gmtime

        log_handler = logging.handlers.WatchedFileHandler(file)
        log_handler.setFormatter(formatter)

        logger = logging.getLogger()
        logger.addHandler(log_handler)
        logger.setLevel(level)

    else:
        # date time in GMT/UTC
        logging.Formatter.converter = time.gmtime
        logging.basicConfig(level=level, format=format, datefmt=datefmt)



import json
import logging
import re
from datetime import date, datetime, time
from enum import Enum
from pathlib import Path

import aiofiles
import tomlkit
import yaml


class FORMAT(Enum):
    TOML = "TOML"
    YAML = "YAML"
    JSON = "JSON"


class Config(dict):
    """Class to have settings in memory or in a configuration file"""

    @classmethod
    def from_file(cls, format: FORMAT, filepath: str):
        """Load a file and return the content in the specified FORMAT

        Arguments
        ---------
        format      FORMAT  A valid config.FORMAT value (TOML, YALM, JSON)
        filepath:   str     The filepath of the file to load
        """
        assert isinstance(format, FORMAT)
        assert isinstance(filepath, str) and len(filepath) > 0

        data = None

        try:
            file = "{path}".format(path=filepath)
            with open(file, "rt", encoding="utf-8") as f:
                if format == FORMAT.TOML:
                    data = tomlkit.load(f)
                elif format == FORMAT.YAML:
                    data = yaml.safe_load(f)
                elif format == FORMAT.JSON:
                    data = json.load(f)

        except Exception as e:
            logger = logging.getLogger(__class__.__name__)
            logger.error("from_file %s error loading %s. %s", format.value, filepath, e)

        return data

    @classmethod
    async def aiofrom_file(cls, format: FORMAT, filepath: str):
        """Load a file in async mode and return the content in the specified FORMAT

        Arguments
        ---------
        format      FORMAT  A valid config.FORMAT value (TOML, YALM, JSON)
        filepath:   str     The filepath of the file to load
        """
        assert isinstance(format, FORMAT)
        assert isinstance(filepath, str) and len(filepath) > 0

        data = None

        try:
            file = "{path}".format(path=filepath)
            async with aiofiles.open(file, "rt", encoding="utf-8") as f:
                if format == FORMAT.TOML:
                    data = tomlkit.loads(await f.read())
                elif format == FORMAT.YAML:
                    data = yaml.safe_load(await f.read())
                elif format == FORMAT.JSON:
                    data = json.loads(await f.read())

        except Exception as e:
            logger = logging.getLogger(__class__.__name__)
            logger.error("aiofrom_file %s error loading %s. %s", format.value, filepath, e)

        return data

    @classmethod
    def from_toml(cls, filepath: str, defaults: dict = None):
        """Load a toml configuration file and return a Config instance

        Arguments
        ---------
        filepath:   str     The filepath of the file to load
        defaults:   dict    A dictionary with default settings.
                            Values from the file will expand/replace defaults
        """
        settings = cls.from_file(FORMAT.TOML, filepath)
        return Config(cls.merge(defaults, settings), filepath)

    @classmethod
    def from_yaml(cls, filepath: str, defaults: dict = None):
        """Load a yaml configuration file and return a Config instance

        Arguments
        ---------
        filepath:   str     The filepath of the file to load
        defaults:   dict    A dictionary with default settings.
                            Values from the file will expand/replace defaults
        """
        settings = cls.from_file(FORMAT.YAML, filepath)
        return Config(cls.merge(defaults, settings), filepath)

    @classmethod
    def from_json(cls, filepath: str, defaults: dict = None):
        """Load a json configuration file and return a Config instance

        Arguments
        ---------
        filepath:   str     The filepath of the file to load
        defaults:   dict    A dictionary with default settings.
                            Values from the file will expand/replace defaults
        """
        settings = cls.from_file(FORMAT.JSON, filepath)
        return Config(cls.merge(defaults, settings), filepath)

    @classmethod
    async def aiofrom_toml(cls, filepath: str, defaults: dict = None):
        """Load a toml configuration file in asycn mode and return a Config instance

        Arguments
        ---------
        filepath:   str     The filepath of the file to load
        defaults:   dict    A dictionary with default settings.
                            Values from the file will expand/replace defaults
        """
        settings = await cls.aiofrom_file(FORMAT.TOML, filepath)
        return Config(cls.merge(defaults, settings), filepath)

    @classmethod
    async def aiofrom_yaml(cls, filepath: str, defaults: dict = None):
        """Load a toml configuration file in asycn mode and return a Config instance

        Arguments
        ---------
        filepath:   str     The filepath of the file to load
        defaults:   dict    A dictionary with default settings.
                            Values from the file will expand/replace defaults
        """
        settings = await cls.aiofrom_file(FORMAT.YAML, filepath)
        return Config(cls.merge(defaults, settings), filepath)

    @classmethod
    async def aiofrom_json(cls, filepath: str, defaults: dict = None):
        """Load a toml configuration file in asycn mode and return a Config instance

        Arguments
        ---------
        filepath:   str     The filepath of the file to load
        defaults:   dict    A dictionary with default settings.
                            Values from the file will expand/replace defaults
        """
        settings = await cls.aiofrom_file(FORMAT.JSON, filepath)
        return Config(cls.merge(defaults, settings), filepath)

    @classmethod
    def merge(cls, defaults: dict = None, settings: dict = None) -> dict:
        """Merge settings into defaults (replace/expand defaults)

        Arguments
        ---------
        defaults:   dict    The default settings
        settings:   dict    The settings to expand/replace into defaults
        """
        assert not defaults or isinstance(defaults, dict)
        assert not settings or isinstance(settings, dict)

        if not defaults:
            defaults = {}
        if not settings:
            settings = {}

        # TODO return a merge including second and nth sub-levels on the dict
        return defaults | settings

    @classmethod
    def to_file(cls, format: FORMAT, filepath: str, data: dict | list):
        """Save data to a file in the specified format

        Arguments
        ---------
        format      FORMAT  A valid config.FORMAT value (TOML, YALM, JSON)
        """
        assert isinstance(format, FORMAT)
        assert isinstance(filepath, str) and len(filepath) > 0
        assert isinstance(data, (dict, list))

        try:
            file = Path(filepath)
            parent = Path(file.parent) if not file.exists() else None
            if parent and not parent.exists():
                parent.mkdir(exist_ok=True, parents=True)

            with open(filepath, "wt", encoding="utf-8") as f:
                if format == FORMAT.TOML:
                    tomlkit.dump(data, f)
                elif format == FORMAT.YAML:
                    yaml.dump(data, f)
                elif format == FORMAT.JSON:
                    json.dump(data, f)

        except Exception as e:
            logger = logging.getLogger(__class__.__name__)
            logger.error("to_file %s error %s. %s", format.value, filepath, e)

    @classmethod
    async def aioto_file(cls, format: FORMAT, filepath: str, data: dict | list):
        """Save data in async mode to a file in the specified format

        Arguments
        ---------
        format      FORMAT  A valid config.FORMAT value (TOML, YALM, JSON)
        """
        assert isinstance(format, FORMAT)
        assert isinstance(filepath, str) and len(filepath) > 0
        assert isinstance(data, (dict, list))

        try:
            file = Path(filepath)
            parent = Path(file.parent) if not file.exists() else None
            if parent and not parent.exists():
                parent.mkdir(exist_ok=True, parents=True)

            async with aiofiles.open(filepath, "wt", encoding="utf-8") as f:
                if format == FORMAT.TOML:
                    await f.write(tomlkit.dumps(data))
                elif format == FORMAT.YAML:
                    await f.write(yaml.dump(data))
                elif format == FORMAT.JSON:
                    await f.write(json.dumps(data))

        except Exception as e:
            logger = logging.getLogger(__class__.__name__)
            logger.error("aioto_file %s error %s. %s", format.value, filepath, e)

    def __init__(self, settings: dict = {}, filepath: str = None):
        """
        Arguments
        ---------
        settings:   dict    A dictionary of settings
                            Each key in the dictionary must start with lowercase a-z
                            and only ASCII characters are allowed in the name [a-ZA-Z_0-9]


        filepath:   str     Full filepath of the file to store settings in
        """
        assert isinstance(settings, dict)
        assert not filepath or isinstance(filepath, str)
        self._logger = logging.getLogger(__class__.__name__)

        self.filepath = filepath

        for key, value in settings.items():
            self.add(key, value)

    def _tofile(self, format: FORMAT):
        """Save settings to file in format

        Arguments
        ---------
        format      FORMAT  A valid config.FORMAT value (TOML, YALM, JSON)
        """
        assert isinstance(format, FORMAT)
        if not self.filepath:
            self._logger.error("_tofile Not filepath defined for %s. Aborting", format.value)
            return

        self.to_file(format=format, filepath=self.filepath, data=self.copy())

    async def _aiotofile(self, format: FORMAT):
        """Save settings to file in format

        Arguments
        ---------
        format      FORMAT  A valid config.FORMAT value (TOML, YALM, JSON)
        """
        assert isinstance(format, FORMAT)

        if not self.filepath:
            self._logger.error("_aiotofile Not filepath defined for %s. Aborting", format.value)
            return

        await self.aioto_file(format=format, filepath=self.filepath, data=self.copy())

    def to_toml(self):
        """Save settings to file in toml format"""
        self._tofile(FORMAT.TOML)

    def to_yaml(self):
        """Save settings to file in yaml format"""
        self._tofile(FORMAT.YAML)

    def to_json(self):
        """Save settings to file in json format"""
        self._tofile(FORMAT.JSON)

    async def aioto_toml(self):
        """Save settings in async mode to file in toml format"""
        await self._aiotofile(FORMAT.TOML)

    async def aioto_yaml(self):
        """Save settings in async mode to file in yaml format"""
        await self._aiotofile(FORMAT.YAML)

    async def aioto_json(self):
        """Save settings in async mode to file in json format"""
        await self._aiotofile(FORMAT.JSON)

    def add(self, key: str, value):
        """Add/set a setting

        Arguments
        ---------
        key:    str         A str valid key to name this setting.
                            The key name must star with a lowercase [a-z], and contain ASCII characters only

        value   object      Value of the setting.  The only allowed values are:
                            str, int, float, list, dict, bool, datetime, date, time
        """
        assert isinstance(key, str) and re.match("[a-z]\w", key)
        assert isinstance(value, (str, int, float, list, dict, bool, datetime, date, time))

        self[key] = value

    def remove(self, key: str):
        """Remove a setting from this configuration

        Arguments
        ---------
        key:    str         The key of the setting to remove
        """
        self.pop(key, None)



import asyncio
import logging
import queue
import random
import signal
import time
from enum import Enum
from multiprocessing import Pool

from minibone.daemon import Daemon


class TypeQueue(Enum):
    """Emum to define Type of queues

    FIFO = First in, first out
    LIFO = Last in, first out
    """

    FIFO = "FIFO"
    LIFO = "LIFO"


class PARProcesses(Daemon):
    """Class to run parallel tasks as processes.

    This is suitable for CPU bound tasks such as heavy calculations and maths
    For I/O bounded tasks use PARThreads instead.

    Usage
    -----
    - Create a PARProcesses instance
    - call start()
    - call queue to add all your heavy computation tasks and store the UID
    - call get(UID) to retrieve the callback's result
    - continue your queue / get loop
    - call stop at the end

    Example
    -------
    from minibone.parallel_processes import PARProcesses

    def mypow(x, y):
        return x ** y

    p = PARProcesses()
    p.start()

    # tasks will be run on its own CPU process
    uid1 = p.queue(mypow, x=5, y=8)
    uid2 = p.queue(mypow, x=8, y=16)

    res1 = p.get(uid1)
    res2 = p.get(uid2)

    p.stop()

    print(res1, res2)
    """

    def __init__(
        self, type_queue: TypeQueue = TypeQueue.FIFO, interval: int = 0, sleep: int = 0.005, daemon: bool = True
    ):
        """
        Arguments
        ---------
        type_queue:     TypeQueue       Type of queue desired (FIFO, LIFO)
        daemon:         bool            True to run as a daemon, False otherwise
        """
        assert isinstance(type_queue, TypeQueue)
        assert isinstance(daemon, bool)
        super().__init__(name="PoolThreads", interval=interval, sleep=sleep, daemon=daemon)

        self._logger = logging.getLogger(__class__.__name__)

        if type_queue == TypeQueue.FIFO:
            self._queue = queue.Queue()
        else:
            self._queue = queue.LifoQueue()

        self._pool = Pool(initializer=self._init_worker)
        self._processing = {}

        random.seed(time.time())
        self._letters = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"

    def _init_worker(self):
        # https://stackoverflow.com/a/6191991
        signal.signal(signal.SIGINT, signal.SIG_IGN)

    def _uid(self, cmd: str) -> str:
        """Return an unique UID

        Arguments
        ---------
        cmd:    str     String to prefix UID
        """
        assert isinstance(cmd, str)
        return "{}{}_{}".format(cmd, self._letters[random.randint(0, len(self._letters) - 1)], time.time())

    def _lookfor_result(self, uid: str) -> object:
        """Search for a result"""
        assert isinstance(uid, str)
        self.lock.acquire()
        resp = self._processing.pop(uid, "_no_in_dict_")
        self.lock.release()
        return resp

    def get(self, uid: str, timeout=-1) -> object | None:
        """Return the response mapped to UID

        Arguments
        ---------
        uid:        str     Unique identifier of the expected result
        timeout:    int     Set the number of seconds to wait for a result.
                            Set to zero or negative to wait forever
        """
        assert isinstance(uid, str)
        assert isinstance(timeout, (int, float))

        notimeout = True if timeout <= 0 else False
        epoch = time.time()
        future_epoch = time.time() + timeout
        while notimeout or epoch < future_epoch:
            resp = self._lookfor_result(uid)
            if resp != "_no_in_dict_" and resp.ready:
                return resp.get()

            time.sleep(0.01)

        return None

    async def aioget(self, uid: str, timeout=-1) -> object | None:
        """Return the response mapped to UID in async mode

        Arguments
        ---------
        uid:        str     Unique identifier of the expected result
        timeout:    int     Set the number of seconds to wait for a result.
                            Set to zero or negative to wait forever
        """
        assert isinstance(uid, str)
        assert isinstance(timeout, (int, float))

        notimeout = True if timeout <= 0 else False
        epoch = time.time()
        future_epoch = time.time() + timeout
        while notimeout or epoch < future_epoch:
            resp = self._lookfor_result(uid)
            if resp != "_no_in_dict_" and resp.ready:
                return resp.get()

            await asyncio.sleep(0.01)

        return None

    def queue(self, callback, cmd: str = "UID", keep: bool = True, **kwargs) -> str:
        """Queue a task and return an UID to get the result later

        Arguments
        ---------
        callback:   object  A callable object
        cmd:        str     Any command to prefix the UID to return
        keep:       bool    Set to True to keep the callback's result in memory
                            Set to False to forget it. Usefull for callbacks that returns nothing
        kwargs:     dict    a dict with key/value parameters

        Notes
        -----
        If keep is set, do not forget to call get(uid) to free memory
        """
        assert callable(callable)
        assert isinstance(keep, bool)
        assert not kwargs or isinstance(kwargs, dict)

        if not kwargs:
            kwargs = dict()

        uid = self._uid(cmd)
        item = {"uid": uid, "callback": callback, "keep": keep, "kwargs": kwargs}
        self._queue.put(item)

        self._logger.debug("Task %s queued", uid)
        return uid

    def on_process(self):
        while self._queue.qsize() > 0:
            item = self._queue.get()
            self._logger.debug("Task %s started", item["uid"])

            resp = self._pool.apply_async(func=item["callback"], kwds=item["kwargs"])

            self.lock.acquire()
            if item["keep"]:
                self._processing[item["uid"]] = resp
            self._queue.task_done()
            self.lock.release()

    def stop(self):
        super().stop()
        self._pool.close()
        self._pool.terminate()
        self._pool.join()



import glob
import logging
import time
from pathlib import Path
from string import Template

import aiofiles

from minibone.config import Config


class HTMLBase:
    """Class to render html using snippets and toml configuration file in async mode"""

    def __init__(self, snippets_path: str = "./html/pages/snippets", ext: str = "html", cache_life: int = 300):
        """
        Arguments
        ---------
        snippets_path:  str     The path where snippets are stored
        ext:            str     The extensio of the snippets files
        cache_life:     int     Number of seconds to have html snippets in memory before reloading them
        """
        assert isinstance(snippets_path, str)
        assert isinstance(ext, str)
        assert isinstance(cache_life, int)
        self._logger = logging.getLogger(__class__.__name__)

        self._snippets_path = snippets_path
        self._ext = ext
        self._cache_life = cache_life
        self._epoch = 0

        self._snippets = {}

    async def _aiofile(self, file: str) -> str:
        """Load content from a file using encoding utf-8 in async mode and return it

        Arguments
        ---------
        file        str     The file to load from disk
        """
        assert isinstance(file, str)
        try:
            async with aiofiles.open(
                file,
                "rt",
                encoding="utf-8",
            ) as f:
                return await f.read()

        except Exception as e:
            self._logger.error("_aiofile %s", e)
            return None

    async def _iosnippets(self) -> str:
        """Load all HTML snippets from files in async mode only when cached is old."""

        epoch = time.time()
        if self._epoch > epoch:
            return

        self._epoch = epoch + self._cache_life

        p = Path(self._snippets_path)
        if p.exists() and p.is_dir():
            files = glob.glob(glob.escape(self._snippets_path) + f"/*.{self._ext}")
            for file in files:
                name = file.split("/")[-1].split(".")[0]
                content = await self._aiofile(file)
                if content:
                    self._snippets[name] = content

    async def aio_file(self, file: str) -> str:
        """Return the file's content using encoding utf-8 in async mode.

        Arguments
        ---------
        file        str     The file to load from disk
        """
        return await self._aiofile(file)

    def render(self, template: str, mapping: dict) -> str:
        """Return a str render for template using mapping keys and values

        Arguments
        ---------
        template    str     The template in string format
        mapping     dict    A dictionary of key an values

        Sample
        ------

        h = HTMLBase()
        template = "<div>Hello ${user}</div>"
        render = h.render(template, {"user": "John"})

        # will return
        # <div>Hello John</div>
        #
        # Further reading in python string.Template module
        """
        assert isinstance(template, str)
        assert isinstance(mapping, dict)

        return Template(template).safe_substitute(mapping)

    async def aiofrom_toml(self, filepath: str) -> str:
        """Load and toml file, extract configuration values then render a html file accordingly in async mode

        Arguments
        ---------
        filepath:       str         The toml's filepath having the html render configuration

        Notes
        -----
        - Minimum toml configuration has a [page] block with 'html_file' settign
        - Add html snippets (.html) into the snippets path
        - In the toml file add additional blocks for each snippet. Named blocks as each snippet file's name
        - Add key/values to be replaced in the snippets (see render method)

        Example
        -------

        1. Into snippets path there is an account.html snippet having next markup

        <div>Hello ${user}</div>


        2. There a file named index.html having:

        <!DOCTYPE html>
        <html lang='es-ES'>
        <head>
            <title>${title}</title>
        </head>
        <body>
            ${account}
        </body>

        3. There is a file named index.toml having next minimal configuration

        [page]
        #
        html_file = 'index.html'
        #
        title = 'Super cool website'

        [account]
        #
        user = John

        4. use next code to get the html rendered

        content = aiofrom_toml("index.toml")
        """
        assert isinstance(filepath, str)

        settings = await Config.aiofrom_toml(filepath=filepath)
        if not settings or not settings.get("page", None):
            self._logger.error("from_toml invalid file %s or not [page] block found", filepath)
            return

        cfg_page = settings["page"]
        if not cfg_page["html_file"]:
            self._logger.error("from_toml file has not html_file setting in [page] block %s", filepath)
            return

        await self._iosnippets()
        for key, snippet in self._snippets.items():
            cfg_page[key] = self.render(snippet, settings.get(key, {}))

        content = await self.aio_file(cfg_page["html_file"])
        content = self.render(content, cfg_page)

        return content



import time
import unittest
from pathlib import Path

from minibone.storing import Storing


class TestStoring(unittest.TestCase):
    def test_settings(self):
        data1 = [1, 4, 5, 6]
        data2 = {"key1": "val1", "key2": "val2", "list": [1, 2, 3]}
        storing = Storing(chunks=2, interval=1)
        storing.to_json(path="./", filename="storing1.json", data=data1)
        storing.to_json(path="./", filename="storing2.json", data=data2)
        storing.start()

        time.sleep(5)

        files = ["./storing1.json", "./storing2.json"]

        json1 = storing.from_json(files[0])
        json2 = storing.from_json(files[1])

        self.assertEqual(json1, [1, 4, 5, 6])
        self.assertEqual(json2, {"key1": "val1", "key2": "val2", "list": [1, 2, 3]})

        storing.stop()

        for file in files:
            p = Path(file)
            p.unlink(missing_ok=True)


if __name__ == "__main__":
    unittest.main()



import time
import unittest
from threading import Lock

from minibone.daemon import Daemon


class DaemonSubClass(Daemon):
    value = 0

    def __init__(self):
        super().__init__(interval=1, iter=3)

    def on_process(self):
        self.lock.acquire()
        self.value += 1
        self.lock.release()


class TestDaemon(unittest.TestCase):
    lock = Lock()
    value = 0

    def callback(self):
        self.lock.acquire()
        self.value += 1
        self.lock.release()

    def test_daemon(self):
        """
        Test Daemon is running in the background doing the job
        """
        task1 = Daemon(interval=1, iter=3, callback=self.callback)
        task2 = Daemon(interval=1, iter=3, callback=self.callback)
        task3 = DaemonSubClass()

        task1.start()
        task2.start()
        task3.start()

        time.sleep(5)
        self.assertEqual(self.value, 6)
        self.assertEqual(task3.value, 3)

        task1.stop()
        task2.stop()
        task3.stop()


if __name__ == "__main__":
    unittest.main()
